# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Read."""

import json


def read_json(filepath):
    """Read json filepath."""
    with open(filepath) as f:
        return json.load(f)


def read_raw_analytics_from_filepaths(filepaths):
    """Iterate (YYYY-MM-DD, analytic) from all filepaths."""
    for fp in filepaths:
        analytics_from_fp = read_json(fp)
        for year_month_day, raw_analytics_list in analytics_from_fp.items():
            for raw_analytics in raw_analytics_list:
                yield (year_month_day, raw_analytics)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Write."""

import json


def write_json(
    filepath, content, sort_keys=True, indent=2, separators=(",", ": "), **kwargs  # noqa
):
    """Write json filepath."""
    with open(filepath, "w") as f:
        json.dump(
            content,
            f,
            sort_keys=sort_keys,
            indent=indent,
            separators=separators,
            **kwargs
        )
    return filepath



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import pytest

from invenio_analytics_importer.cache import (
    fill_downloads_cache,
    fill_views_cache,
)
from invenio_analytics_importer.convert import DownloadAnalytics, ViewAnalytics


def test_fill_downloads_cache(running_app, db, record_factory):
    file_key = "coffee.assess.nobmi.txt"
    r = record_factory.create_record(filenames=[file_key])
    r.index.refresh()
    pid = r.pid.pid_value
    iter_analytics = [
        DownloadAnalytics(
            year_month_day="2024-09-03",
            pid=pid,
            file_key=file_key,
            visits=2,
            views=3,
        ),
        DownloadAnalytics(
            year_month_day="2024-09-03",
            pid="notin-cache",
            file_key="doesnt_exist.txt",
            visits=1,
            views=1,
        ),
    ]

    cache = fill_downloads_cache(iter_analytics)

    file_model = r.files[file_key].file.file_model
    file_id = cache.get_file_id(pid, file_key)
    assert str(file_model.id) == file_id
    assert r.parent.pid.pid_value == cache.get_parent_pid(pid)
    assert file_model.size == cache.get_size(file_id)
    assert str(r.bucket_id) == cache.get_bucket_id(pid)

    with pytest.raises(KeyError):
        cache.get_parent_pid("notin-cache")
    with pytest.raises(KeyError):
        cache.get_file_id("notin-cache", "doesnt_exist.txt")


def test_fill_views_cache(running_app, db, record_factory):
    r = record_factory.create_record()
    r.index.refresh()
    pid = r.pid.pid_value
    iter_analytics = [
        ViewAnalytics(
            year_month_day="2024-09-03",
            pid=pid,
            visits=2,
            views=3,
        ),
        ViewAnalytics(
            year_month_day="2024-09-03",
            pid="notin-cache",
            visits=1,
            views=1,
        ),
    ]

    cache = fill_views_cache(iter_analytics)

    assert r.parent.pid.pid_value == cache.get_parent_pid(pid)
    with pytest.raises(KeyError):
        cache.get_parent_pid("notin-cache")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Config."""

ANALYTICS_IMPORTER_MATOMO_URL = ""
"""Your Matamo site's base url."""

ANALYTICS_IMPORTER_MATOMO_SITE_ID = ""
"""Your Matamo site's id."""

ANALYTICS_IMPORTER_MATOMO_TOKEN = ""
"""A Matamoto application token."""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Retrieve aggregate analytics from provider."""

import abc
import asyncio
import calendar
import dataclasses
from typing import Any

import httpx
from flask import current_app

from invenio_analytics_importer.write import write_json


class ProviderClient(abc.ABC):
    """Provider client interface."""

    @abc.abstractmethod
    async def fetch_views_for_day(self, day):
        """Fetch views for given day."""

    @abc.abstractmethod
    async def fetch_downloads_for_day(self, day):
        """Fetch downloads for given day."""


@dataclasses.dataclass
class MatomoAnalytics(ProviderClient):
    """Matomo API client."""

    client: Any
    base_url: str
    site_id: int
    token_auth: str

    async def get_analytics_for_day(self, method, day):
        """Get analytics for given YYYY-MM-DD in json format.

        Response format:

        {
            "label": "example.org/records/0dfv3-cmw61/files/f.pdf?download=1]",
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
        },
        """
        params = {
            "module": "API",
            "format": "json",
            "idSite": self.site_id,
            "period": "day",
            "date": day,
            "method": method,
            "flat": 1,
            "showMetadata": 0,
        }

        try:
            response = await self.client.post(
                self.base_url,
                params=params,
                data={"token_auth": self.token_auth}
            )
            response.raise_for_status()
        except httpx.HTTPError as exc:
            print(f"Error : {exc.request.url} : {exc}")
            return []

        if response.text == "No data available":
            return []

        return response.json()

    async def fetch_downloads_for_day(self, day):
        """Fetch downloads analytics for day."""
        return await self.get_analytics_for_day("Actions.getDownloads", day)

    async def fetch_views_for_day(self, day):
        """Fetch views analytics for day."""
        return await self.get_analytics_for_day("Actions.getPageUrls", day)


class ViewsFetcher:
    """Fetches views."""

    def __init__(self, client):
        """Constructor."""
        self.client = client

    async def fetch_analytics_for_day(self, day):
        """Fetch views analytics for given day."""
        return await self.client.fetch_views_for_day(day)


class DownloadsFetcher:
    """Fetches downloads."""

    def __init__(self, client):
        """Constructor."""
        self.client = client

    async def fetch_analytics_for_day(self, day):
        """Fetch download analytics for given day."""
        return await self.client.fetch_downloads_for_day(day)


def generate_days_by_year_month(period):
    """
    Yield (YYYY-MM, (YYYY-MM-DD1, ..., YYYY-MM-DDN)) for each month of period.

    :param period: tuple. (date_from, date_to)
        :param date_from: str. YYYY-MM-DD from inclusive.
        :param date_to: str. YYYY-MM-DD to inclusive.
    """
    date_from, date_to = period
    tmp = date_from.split("-")
    year_from, month_from = int(tmp[0]), int(tmp[1])
    tmp = date_to.split("-")
    year_to, month_to = int(tmp[0]), int(tmp[1])

    y = year_from
    m = month_from
    while (y, m) <= (year_to, month_to):
        days_in_month = calendar.monthrange(y, m)[1]
        year_month = f"{y:4}-{m:02}"
        days = (
            f"{y:4}-{m:02}-{d:02}" for d in range(1, days_in_month + 1)
        )

        yield year_month, days

        m += 1
        if m > 12:
            y += 1
            m = 1


async def fetch_monthly_analytics(fetcher, period):
    """Yield fetched analytics by month (and day within month)."""
    days_by_yr_m = generate_days_by_year_month(period)

    for year_month, days in days_by_yr_m:
        days_l = list(days)
        # This maps to 30~ concurrent network calls at a time which has been
        # fine so far. It may return lots of data to hold in memory though.
        # This has been fine for us so far as well, but it's a trade-off in
        # favor of less files and more understandable periods.
        analytics_daily = await asyncio.gather(
            *[fetcher.fetch_analytics_for_day(day) for day in days_l]
        )
        analytics_monthly = dict(zip(days_l, analytics_daily))
        yield year_month, analytics_monthly


async def retrieve_period_analytics(provider, kind, period, output_dir):
    """Framing device."""
    async with httpx.AsyncClient() as client:
        # If other providers, do selection + creation here.
        # For now, there isn't, so no selection logic.
        client_of_provider = MatomoAnalytics(
            client,
            base_url=current_app.config.get("ANALYTICS_IMPORTER_MATOMO_URL"),
            site_id=current_app.config.get("ANALYTICS_IMPORTER_MATOMO_SITE_ID"),  # noqa
            token_auth=current_app.config.get("ANALYTICS_IMPORTER_MATOMO_TOKEN"),  # noqa
        )

        if kind == "views":
            fetcher = ViewsFetcher(client_of_provider)
        elif kind == "downloads":
            fetcher = DownloadsFetcher(client_of_provider)
        else:
            exit(1)

        async for yr_m, analytics in fetch_monthly_analytics(fetcher, period):
            write_json(output_dir / f"{kind}_{yr_m}.json", analytics)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import dataclasses
import json

import pytest

from invenio_analytics_importer.retrieve import (
    DownloadsFetcher,
    MatomoAnalytics,
    ProviderClient,
    ViewsFetcher,
    fetch_monthly_analytics,
)


@dataclasses.dataclass
class FakeResponse:
    """Fake response."""

    json_body: str

    def raise_for_status(self):
        """Faked raise_for_status."""
        return False

    @property
    def text(self):
        """Faked text."""
        return self.json_body

    def json(self):
        """Faked json."""
        return json.loads(self.json_body)


class FakeClient:
    """Fake httpx client."""

    def __init__(self):
        """Constructor."""
        self._response_data = {}

    def set_response(self, criteria, json_body=""):
        """Register a response."""
        self._response_data[criteria] = {
            "json_body": json_body,
        }

    def get_response(self, criteria):
        """Get response."""
        data = self._response_data[criteria]
        return FakeResponse(**data)

    async def post(self, url, params=None, data=None):
        """Post."""
        return self.get_response((url, params["date"]))


@pytest.mark.asyncio
async def test_matomo_get_analytics_for_day(running_app):
    analytics_2024_08_01 = [
        {
            "label": f"prism.northwestern.edu/records/paah4-s0w35/files/PNB-7-75.txt?download=1",  # noqa
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
            "sum_time_spent": 0,
        },
        {
            "label": f"prism.northwestern.edu/records/t8k1h-p8435/files/PNB 7 76.txt?download=1",  # noqa
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
            "sum_time_spent": 0,
        },
    ]
    fake_client = FakeClient()
    base_url = "https://matomo.example.org/"
    fake_client.set_response(
        (base_url, "2024-08-01"),
        json.dumps(analytics_2024_08_01)
    )
    fake_client.set_response((base_url, "2024-08-02"), "No data available")
    site_id = 3
    token = "token"
    matomo = MatomoAnalytics(fake_client, base_url, site_id, token)

    # Test with data
    result = await matomo.get_analytics_for_day("aMethod", "2024-08-01")

    assert analytics_2024_08_01 == result

    # Test when no data available
    result = await matomo.get_analytics_for_day("aMethod", "2024-08-02")

    assert [] == result


class FakeProviderClient(ProviderClient):
    """Fave provider client."""

    async def fetch_downloads_for_day(self, day):
        """Fetch downloads for day."""
        return [{"downloads": 1}]

    async def fetch_views_for_day(self, day):
        """Fetch views for day."""
        return [{"views": 1}]


@pytest.mark.asyncio
async def test_downloads_fetcher():
    # We are simply testing interfaces here
    fetcher = DownloadsFetcher(FakeProviderClient())

    results = await fetcher.fetch_analytics_for_day("2024-08-31")

    assert [{"downloads": 1}] == results


@pytest.mark.asyncio
async def test_views_fetcher():
    # We are simply testing interfaces here
    fetcher = ViewsFetcher(FakeProviderClient())

    results = await fetcher.fetch_analytics_for_day("2024-08-31")

    assert [{"views": 1}] == results


class FakeFetcher:
    """Fetches downloads."""

    def __init__(self):
        """Constructor."""
        self._response_data = {}

    def set_response(self, day, json_list):
        """Register a response."""
        self._response_data[day] = json_list

    async def fetch_analytics_for_day(self, day):
        """Fetch download analytics for given day."""
        return self._response_data.get(day, [])


@pytest.mark.asyncio
async def test_fetch_monthly_analytics():
    fetcher = FakeFetcher()
    analytics_2023_12_01 = [
        {
            "label": f"prism.northwestern.edu/records/paah4-s0w35/files/PNB-7-75.txt?download=1",  # noqa
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
            "sum_time_spent": 0,
        },
        {
            "label": f"prism.northwestern.edu/records/t8k1h-p8435/files/PNB 7 76.txt?download=1",  # noqa
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
            "sum_time_spent": 0,
        },
    ]
    analytics_2024_01_31 = [
        {
            "label": f"/files/PNB 7 76.txt?download=1",
            "nb_hits": 4,
            "nb_uniq_visitors": 3,
            "nb_visits": 2,
            "sum_time_spent": 0,
        },
    ]
    # only fill 2 days
    fetcher.set_response("2023-12-01", analytics_2023_12_01)
    fetcher.set_response("2024-01-31", analytics_2024_01_31)
    period = ("2023-12", "2024-01")

    monthly_analytics = []
    async for e in fetch_monthly_analytics(fetcher, period):
        monthly_analytics.append(e)

    assert 2 == len(monthly_analytics)

    month, analytics = monthly_analytics[0]
    assert "2023-12" == month
    assert analytics_2023_12_01 == analytics["2023-12-01"]
    # non filled should just be []
    # just take 1 random entry
    assert [] == analytics["2023-12-28"]

    month, analytics = monthly_analytics[1]
    assert "2024-01" == month
    assert analytics_2024_01_31 == analytics["2024-01-31"]
    # non filled should just be []
    # just take 1 random entry
    assert [] == analytics["2024-01-02"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

from invenio_analytics_importer.read import read_raw_analytics_from_filepaths
from invenio_analytics_importer.write import write_json


def test_read_raw_analytics_from_filepaths(tmp_path):
    download_dict_2024_08_31_2 = {
        "label": f"prism.northwestern.edu/records/3s45v-k5m55/files/PNB 7 76.txt?download=1",  # noqa
        "nb_hits": 1,
        "nb_uniq_visitors": 1,
        "nb_visits": 1,
        "sum_time_spent": 0,
    }

    fp1 = write_json(
        tmp_path / "downloads_2024-08.json",
        {
            "2024-08-31": [
                {
                    "label": f"prism.northwestern.edu/records/3s45v-k5m55/files/PNB-7-75.txt?download=1",  # noqa
                    "nb_hits": 1,
                    "nb_uniq_visitors": 1,
                    "nb_visits": 1,
                    "sum_time_spent": 0,
                },
                download_dict_2024_08_31_2,
            ]
        }
    )
    view_dict_2024_08_30 = {
        "avg_page_load_time": 1.904,
        "avg_time_dom_completion": 0.629,
        "avg_time_dom_processing": 0.412,
        "avg_time_network": 0.473,
        "avg_time_on_page": 454,
        "avg_time_server": 0.294,
        "avg_time_transfer": 0.096,
        "bounce_rate": "100%",
        "entry_bounce_count": "1",
        "entry_nb_actions": "1",
        "entry_nb_uniq_visitors": 1,
        "entry_nb_visits": "1",
        "entry_sum_visit_length": "0",
        "exit_nb_uniq_visitors": 1,
        "exit_nb_visits": "2",
        "exit_rate": "67%",
        "label": "/records/3s45v-k5m55",
        "max_time_dom_completion": "0.6440",
        "max_time_dom_processing": "0.7940",
        "max_time_network": "1.3810",
        "max_time_server": "0.4350",
        "max_time_transfer": "0.1760",
        "min_time_dom_completion": "0.6130",
        "min_time_dom_processing": "0.2180",
        "min_time_network": "0.0000",
        "min_time_server": "0.0020",
        "min_time_transfer": "0.0020",
        "nb_hits": 4,
        "nb_hits_following_search": "3",
        "nb_hits_with_time_dom_completion": "2",
        "nb_hits_with_time_dom_processing": "4",
        "nb_hits_with_time_network": "4",
        "nb_hits_with_time_server": "4",
        "nb_hits_with_time_transfer": "4",
        "nb_uniq_visitors": 1,
        "nb_visits": 3,
        "sum_time_spent": 1817,
    }
    fp2 = write_json(
        tmp_path / "views_2024-08.json", {"2024-08-30": [view_dict_2024_08_30]}
    )
    filepaths = [fp1, fp2]

    raw_analytics = list(read_raw_analytics_from_filepaths(filepaths))

    assert 3 == len(raw_analytics)
    # just check the 2nd and 3rd
    assert "2024-08-31" == raw_analytics[1][0]
    assert download_dict_2024_08_31_2 == raw_analytics[1][1]
    assert "2024-08-30" == raw_analytics[2][0]
    assert view_dict_2024_08_30 == raw_analytics[2][1]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Convert data from one format to another.

WARNING: all of the below assumes Matomo output for now.
"""

import dataclasses
import re

REGEX_PID = re.compile(r"/records/([^/]*)(?:/|$)")
REGEX_DOWNLOAD = re.compile(r"\?download=1$")


def is_record(analytics_raw):
    """Return if analytics is for a record."""
    label = analytics_raw.get("label", "")
    match = REGEX_PID.search(label)
    return bool(match.group(1)) if match else False


def is_download(analytics_raw):
    """Return if analytics is for a download.

    Although Matomo returns what it considers a download,
    this isn't always what we consider a download. For the purposes of
    InvenioRDM, a download is what Matomo returns + filtering on presence
    of `download=1` querystring parameter (i.e. not a preview, and always
    unique/at end of label).
    """
    label = analytics_raw.get("label", "")
    match = REGEX_DOWNLOAD.search(label)
    return bool(match)


@dataclasses.dataclass
class DownloadAnalytics:
    """Intermediate representation."""

    year_month_day: str  # keeping it simple for now
    pid: str
    file_key: str
    visits: int
    views: int

    @classmethod
    def create(cls, year_month_day, analytics_raw):
        """Create Entry from raw analytics."""
        label = analytics_raw.get("label", "")

        # extract "3s45v-k5m55" from ".../records/3s45v-k5m55[/...]"
        # assumes is_record was run on analytics_raw prior to this point
        pid = REGEX_PID.search(label).group(1)

        # extract file key
        regex_key = re.compile(r"/files/([^?]*)\?download=1")
        file_key = regex_key.search(label).group(1)

        return cls(
            year_month_day=year_month_day,
            pid=pid,
            file_key=file_key,
            visits=analytics_raw.get("nb_visits", 0),
            views=analytics_raw.get("nb_hits", 0),
        )


def generate_download_analytics(raw_analytics):
    """Yield DownloadAnalytics entries from raw entries."""
    for year_month_day, raw in raw_analytics:
        if is_record(raw) and is_download(raw):
            yield DownloadAnalytics.create(year_month_day, raw)


@dataclasses.dataclass
class ViewAnalytics:
    """Intermediate representation of view analytics."""

    year_month_day: str  # keeping it simple for now
    pid: str
    visits: int
    views: int

    @classmethod
    def create(cls, year_month_day, analytics_raw):
        """Create from raw analytics."""
        label = analytics_raw.get("label", "")

        # extract "3s45v-k5m55" from ".../records/3s45v-k5m55[/...]"
        # assumes is_record was run on analytics_raw prior to this point
        pid = REGEX_PID.search(label).group(1)

        return cls(
            year_month_day=year_month_day,
            pid=pid,
            visits=analytics_raw.get("nb_visits", 0),
            views=analytics_raw.get("nb_hits", 0),
        )


def generate_view_analytics(raw_analytics):
    """Yield ViewAnalytics entries from raw entries."""
    for year_month_day, raw in raw_analytics:
        if is_record(raw):
            yield ViewAnalytics.create(year_month_day, raw)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Cache of data needed for ingestion work."""

from invenio_db import db
from invenio_rdm_records.records import RDMRecord
from invenio_search import current_search_client
from invenio_search.engine import dsl
from sqlalchemy import select


def scan_records(record_cls, pids):
    """Retrieve SE record."""
    s = dsl.Search(
        using=current_search_client,
        index=record_cls.index.search_alias
    )

    s = s.filter({"terms": {"id": pids}})

    response = s.scan()

    return response


class Cache:
    """Cache of system info."""

    def __init__(self):
        """Constructor."""
        self._data = []
        self._pid_record_and_file_key_to_file_id = {}
        self._pid_record_to_bucket_id = {}
        self._file_id_to_size = {}
        self._pid_record_to_pid_parent = {}

    def get_file_id(self, pid_of_record, file_key):
        """Get file id."""
        return self._pid_record_and_file_key_to_file_id[(pid_of_record, file_key)]  # noqa

    def set_file_id(self, pid_of_record, file_key, file_id):
        """Set file id."""
        self._pid_record_and_file_key_to_file_id[(pid_of_record, file_key)] = file_id  # noqa

    def get_bucket_id(self, pid_of_record):
        """Get bucket id associated with record."""
        return self._pid_record_to_bucket_id[pid_of_record]

    def set_bucket_id(self, pid_of_record, bucket_id):
        """Set bucket id associated with record."""
        self._pid_record_to_bucket_id[pid_of_record] = str(bucket_id)

    def get_size(self, file_id):
        """Get size in bytes of file."""
        return self._file_id_to_size.get(file_id, 0)

    def set_size(self, file_id, size):
        """Set size in bytes of file."""
        self._file_id_to_size[file_id] = size

    def get_parent_pid(self, pid_of_record):
        """Get pid of parent of record."""
        return self._pid_record_to_pid_parent[pid_of_record]

    def set_parent_pid(self, pid_of_record, pid_of_parent):
        """Set pid of parent of record."""
        self._pid_record_to_pid_parent[pid_of_record] = pid_of_parent


def fill_downloads_cache(analytics):
    """Fill and return a downloads cache."""
    cache = Cache()

    pids_set = set(a.pid for a in analytics)

    record_cls = RDMRecord
    uuid_to_pid = {}

    # Get and populate file_ids, size, and parent_id
    results = scan_records(record_cls, list(pids_set))
    for hit in results:
        pid = hit["id"]
        pid_of_parent = hit["parent"]["id"]
        uuid = hit["uuid"]

        cache.set_parent_pid(pid, pid_of_parent)
        uuid_to_pid[uuid] = pid

        entries = hit.get("files", {}).get("entries", [])
        for e in entries:
            key = e["key"]
            file_id = e["file_id"]
            size = e["size"]

            cache.set_file_id(pid, key, file_id)
            cache.set_size(file_id, size)

    # Get and populate bucket id
    stmt = select(
        record_cls.model_cls.id,
        record_cls.model_cls.bucket_id,
    ).where(record_cls.model_cls.id.in_(list(uuid_to_pid.keys())))

    for uuid_of_record, bucket_id in db.session.execute(stmt):
        pid = uuid_to_pid[str(uuid_of_record)]
        cache.set_bucket_id(pid, bucket_id)

    return cache


def fill_views_cache(analytics):
    """Fill and return a views cache."""
    cache = Cache()

    pids_set = set(a.pid for a in analytics)

    # Get and populate parent_id
    record_cls = RDMRecord
    results = scan_records(record_cls, list(pids_set))
    for hit in results:
        pid = hit["id"]
        pid_of_parent = hit["parent"]["id"]
        cache.set_parent_pid(pid, pid_of_parent)

    return cache



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import datetime as dt

import time_machine
from invenio_search import current_search_client
from invenio_search.engine import dsl

from invenio_analytics_importer.cache import Cache
from invenio_analytics_importer.convert import DownloadAnalytics, ViewAnalytics
from invenio_analytics_importer.ingest import (
    generate_download_stats,
    generate_view_stats,
    ingest_statistics,
)


def test_generate_download_stats():
    cache = Cache()
    cache.set_size("cb297587-b25c-4675-832a-d5d2634551c7", 9)
    cache.set_file_id(
        "5ret9-dwz86",
        file_key="coffee.assess.nobmi.txt",
        file_id="cb297587-b25c-4675-832a-d5d2634551c7",
    )
    cache.set_bucket_id("5ret9-dwz86", "1741e723-420a-4023-ad89-7aedebab7bb1")
    cache.set_parent_pid("5ret9-dwz86", "tb2gj-axd97")

    iter_analytics = [
        DownloadAnalytics(
            year_month_day="2024-09-03",
            pid="5ret9-dwz86",
            file_key="coffee.assess.nobmi.txt",
            visits=2,
            views=3,
        ),
        DownloadAnalytics(
            year_month_day="2024-09-03",
            pid="notin-cache",
            file_key="doesnt_exist.txt",
            visits=1,
            views=1,
        ),
        DownloadAnalytics(
            year_month_day="2024-09-03",
            pid="5ret9-dwz86",  # in cache
            file_key="doesnt_exist.txt",  # not in cache
            visits=2,
            views=2,
        ),
    ]

    pit = dt.datetime(2025, 9, 23, 0, 0, 0, tzinfo=dt.timezone.utc)
    with time_machine.travel(pit):
        stats = generate_download_stats(iter_analytics, cache)
        stats = list(stats)

    expected = {
        "_id": "1741e723-420a-4023-ad89-7aedebab7bb1_cb297587-b25c-4675-832a-d5d2634551c7-2024-09-03",  # noqa
        "_index": f"stats-file-download-2024-09",
        "_source": {
            "timestamp": f"2024-09-03T00:00:00",
            "unique_id": "1741e723-420a-4023-ad89-7aedebab7bb1_cb297587-b25c-4675-832a-d5d2634551c7",  # noqa
            "count": 3,
            "updated_timestamp": "2025-09-23T00:00:00+00:00",
            "unique_count": 2,
            "volume": 27,
            "file_id": "cb297587-b25c-4675-832a-d5d2634551c7",
            "file_key": "coffee.assess.nobmi.txt",
            "bucket_id": "1741e723-420a-4023-ad89-7aedebab7bb1",
            "recid": "5ret9-dwz86",
            "parent_recid": "tb2gj-axd97",
        },
    }
    assert 1 == len(stats)
    assert expected == stats[0]


def test_generate_view_stats():
    cache = Cache()
    cache.set_parent_pid("0c8rx-zsn76", "by5n4-x1h80")

    iter_analytics = [
        ViewAnalytics(
            year_month_day="2024-08-30",
            pid="0c8rx-zsn76",
            visits=2,
            views=3,
        ),
        ViewAnalytics(
            year_month_day="2024-09-23",
            pid="notin-cache",
            visits=1,
            views=1,
        ),
    ]

    pit = dt.datetime(2025, 9, 23, 0, 0, 0, tzinfo=dt.timezone.utc)
    with time_machine.travel(pit):
        stats = list(generate_view_stats(iter_analytics, cache))

    expected = {
        "_id": f"ui_0c8rx-zsn76-2024-08-30",
        "_index": f"stats-record-view-2024-08",
        "_source": {
            "timestamp": f"2024-08-30T00:00:00",
            "unique_id": "ui_0c8rx-zsn76",
            "count": 3,
            "updated_timestamp": "2025-09-23T00:00:00+00:00",
            "unique_count": 2,
            "recid": "0c8rx-zsn76",
            "parent_recid": "by5n4-x1h80",
            "via_api": False,
        },
    }
    assert 1 == len(stats)
    assert expected == stats[0]


def test_ingest_statistics(running_app):
    stats_for_ingest = [
        {
            "_id": f"ui_0c8rx-zsn76-2024-08-30",
            "_index": f"stats-record-view-2024-08",
            "_source": {
                "timestamp": f"2024-08-30T00:00:00",
                "unique_id": "ui_0c8rx-zsn76",
                "count": 1,
                "updated_timestamp": dt.datetime.now(dt.timezone.utc).isoformat(),  # noqa
                "unique_count": 1,
                "recid": "0c8rx-zsn76",
                "parent_recid": "by5n4-x1h80",
                "via_api": False,
            },
        },
        {
            "_id": "1741e723-420a-4023-ad89-7aedebab7bb1_cb297587-b25c-4675-832a-d5d2634551c7-2024-09-03",  # noqa
            "_index": f"stats-file-download-2024-09",
            "_source": {
                "timestamp": f"2024-09-03T00:00:00",
                "unique_id": "1741e723-420a-4023-ad89-7aedebab7bb1_cb297587-b25c-4675-832a-d5d2634551c7",  # noqa
                "count": 3,
                "updated_timestamp": dt.datetime.now(dt.timezone.utc).isoformat(),  # noqa
                "unique_count": 2,
                "volume": 27,
                "file_id": "cb297587-b25c-4675-832a-d5d2634551c7",
                "file_key": "coffee.assess.nobmi.txt",
                "bucket_id": "1741e723-420a-4023-ad89-7aedebab7bb1",
                "recid": "5ret9-dwz86",
                "parent_recid": "tb2gj-axd97",
            },
        },
    ]

    # Actual function under test
    ingest_statistics(stats_for_ingest)
    current_search_client.indices.refresh(index="stats-file-*")
    current_search_client.indices.refresh(index="stats-record-*")

    s = dsl.Search(
        using=current_search_client,
        index="stats-record-view-2024-08"
    )
    unique_id = "ui_0c8rx-zsn76"
    result = next(
        (h for h in s.scan() if h["unique_id"] == unique_id),
        None
    )
    assert "timestamp" in result
    assert 1 == result["count"]
    assert "updated_timestamp" in result
    assert 1 == result["unique_count"]
    assert "0c8rx-zsn76" == result["recid"]
    assert "by5n4-x1h80" == result["parent_recid"]
    assert result["via_api"] is False

    s = dsl.Search(
        using=current_search_client,
        index="stats-file-download-2024-09"
    )
    unique_id = "1741e723-420a-4023-ad89-7aedebab7bb1_cb297587-b25c-4675-832a-d5d2634551c7"  # noqa
    result = next(
        (
            h for h in s.scan()
            if h["unique_id"] == unique_id
        ),
        None
    )
    assert "timestamp" in result
    assert 3 == result["count"]
    assert "updated_timestamp" in result
    assert 2 == result["unique_count"]
    assert 27 == result["volume"]
    assert "cb297587-b25c-4675-832a-d5d2634551c7" == result["file_id"]
    assert "coffee.assess.nobmi.txt" == result["file_key"]
    assert "1741e723-420a-4023-ad89-7aedebab7bb1" == result["bucket_id"]
    assert "5ret9-dwz86" == result["recid"]
    assert "tb2gj-axd97" == result["parent_recid"]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

import pytest

from invenio_analytics_importer.convert import (
    generate_download_analytics,
    generate_view_analytics,
)


@pytest.fixture
def download_analytics_for_non_record_raw():
    """Raw Matamo analytics entry."""
    return (
        "2024-08-30",
        {
            "label": "example.org/coffee.assess.bmi.gz?download=1",
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
            "sum_time_spent": 0,
        },
    )


@pytest.fixture
def download_analytics_raw():
    """Raw Matamo analytics entry."""
    return (
        "2024-08-30",
        {
            "label": "example.org/records/3s45v-k5m55/files/coffee.assess.bmi.gz?download=1",  # noqa
            "nb_hits": 5,
            "nb_uniq_visitors": 2,
            "nb_visits": 3,
            "sum_time_spent": 0,
        },
    )


@pytest.fixture
def download_analytics_for_non_download_raw():
    """Raw non-download analytics."""
    return (
        "2024-01-10",
        {
            "label": "example.org/records/y4dqk-7yj30/preview/Bachelors_Degree_1_a_preservation.wav",  # noqa
            "nb_hits": 1,
            "nb_uniq_visitors": 1,
            "nb_visits": 1,
            "sum_time_spent": 0
        }
    )


def test_generate_download_analytics(
    download_analytics_for_non_record_raw,
    download_analytics_raw,
    download_analytics_for_non_download_raw,
):
    analytics = list(
        generate_download_analytics(
            [
                download_analytics_raw,
                download_analytics_for_non_record_raw,
                download_analytics_for_non_download_raw,
            ]
        )
    )

    assert 1 == len(analytics)
    entry = analytics[0]
    assert "3s45v-k5m55" == entry.pid
    assert "coffee.assess.bmi.gz" == entry.file_key
    assert 3 == entry.visits
    assert 5 == entry.views
    assert "2024-08-30" == entry.year_month_day


@pytest.fixture
def view_analytics_for_root_raw():
    return (
        "2025-08-29",
        {
            "avg_page_load_time": 0.578,
            "avg_time_dom_completion": 0,
            "avg_time_dom_processing": 0.476,
            "avg_time_network": 0.067,
            "avg_time_on_page": 110,
            "avg_time_server": 0.034,
            "avg_time_transfer": 0.001,
            "bounce_rate": "0%",
            "entry_bounce_count": "0",
            "entry_nb_actions": "22",
            "entry_nb_uniq_visitors": "3",
            "entry_nb_visits": "4",
            "entry_sum_visit_length": "1147",
            "exit_nb_uniq_visitors": "2",
            "exit_nb_visits": "2",
            "exit_rate": "50%",
            "label": "/",
            "max_time_dom_completion": None,
            "max_time_dom_processing": "1.0880",
            "max_time_network": "0.4890",
            "max_time_server": "0.0870",
            "max_time_transfer": "0.0020",
            "min_time_dom_completion": None,
            "min_time_dom_processing": "0.0670",
            "min_time_network": "0.0000",
            "min_time_server": "0.0010",
            "min_time_transfer": "0.0000",
            "nb_hits": 10,
            "nb_hits_with_time_dom_completion": "0",
            "nb_hits_with_time_dom_processing": "10",
            "nb_hits_with_time_network": "10",
            "nb_hits_with_time_server": "10",
            "nb_hits_with_time_transfer": "10",
            "nb_uniq_visitors": 3,
            "nb_visits": 4,
            "sum_time_spent": 1103,
        },
    )


@pytest.fixture
def view_analytics_raw():
    """Raw Matamo view analytics entry."""
    return (
        "2024-08-30",
        {
            "avg_page_load_time": 1.904,
            "avg_time_dom_completion": 0.629,
            "avg_time_dom_processing": 0.412,
            "avg_time_network": 0.473,
            "avg_time_on_page": 454,
            "avg_time_server": 0.294,
            "avg_time_transfer": 0.096,
            "bounce_rate": "100%",
            "entry_bounce_count": "1",
            "entry_nb_actions": "1",
            "entry_nb_uniq_visitors": 1,
            "entry_nb_visits": "1",
            "entry_sum_visit_length": "0",
            "exit_nb_uniq_visitors": 1,
            "exit_nb_visits": "2",
            "exit_rate": "67%",
            "label": "/records/3s45v-k5m55",
            "max_time_dom_completion": "0.6440",
            "max_time_dom_processing": "0.7940",
            "max_time_network": "1.3810",
            "max_time_server": "0.4350",
            "max_time_transfer": "0.1760",
            "min_time_dom_completion": "0.6130",
            "min_time_dom_processing": "0.2180",
            "min_time_network": "0.0000",
            "min_time_server": "0.0020",
            "min_time_transfer": "0.0020",
            "nb_hits": 4,
            "nb_hits_following_search": "3",
            "nb_hits_with_time_dom_completion": "2",
            "nb_hits_with_time_dom_processing": "4",
            "nb_hits_with_time_network": "4",
            "nb_hits_with_time_server": "4",
            "nb_hits_with_time_transfer": "4",
            "nb_uniq_visitors": 1,
            "nb_visits": 3,
            "sum_time_spent": 1817,
        },
    )


def test_generate_view_analytics(
    view_analytics_for_root_raw, view_analytics_raw
):
    analytics = list(
        generate_view_analytics(
            [
                view_analytics_for_root_raw,
                view_analytics_raw,
            ]
        )
    )

    assert 1 == len(analytics)
    entry = analytics[0]
    assert "3s45v-k5m55" == entry.pid
    assert 3 == entry.visits
    assert 4 == entry.views
    assert "2024-08-30" == entry.year_month_day



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Ingest analytics into InvenioRDM's stats indices."""

from datetime import datetime, timezone

from invenio_search import current_search_client
from invenio_search.engine import search


def record_exists(entry, cache):
    """Checks if the record of passed analytic exists.

    Really this checks if the analytic is for a record in the system and uses
    the cache as the proxy for that.
    """
    try:
        # This would work for views or downloads, so is used to assess presence
        # in cache (assumes some knowledge of how cache is filled)
        return cache.get_parent_pid(entry.pid) is not None
    except KeyError:
        return False


def file_key_exists(entry, cache):
    """Checks if the file key of passed analytic exists.

    Really this checks if the file of the record exists in the system and uses
    the cache as the proxy for that.
    """
    try:
        return cache.get_file_id(entry.pid, entry.file_key) is not None
    except KeyError:
        return False


def to_download(entry, cache):
    """To download."""
    file_id = cache.get_file_id(entry.pid, entry.file_key)
    bucket_id = cache.get_bucket_id(entry.pid)
    year_month_day = entry.year_month_day
    event_name = "file-download"
    year_month = year_month_day.rsplit("-", 1)[0]
    count = entry.views
    unique_count = entry.visits
    volume = count * cache.get_size(file_id)
    file_key = entry.file_key
    recid = entry.pid
    parent_recid = cache.get_parent_pid(recid)

    return {
        "_id": f"{bucket_id}_{file_id}-{year_month_day}",
        "_index": f"stats-{event_name}-{year_month}",
        "_source": {
            # Since those entries are synthetic anyway, we place them at
            # the start of the day
            "timestamp": f"{year_month_day}T00:00:00",
            "unique_id": f"{bucket_id}_{file_id}",
            "count": count,
            "updated_timestamp": datetime.now(timezone.utc).isoformat(),
            "unique_count": unique_count,
            "volume": volume,
            "file_key": file_key,
            "bucket_id": bucket_id,
            "file_id": file_id,
            "recid": recid,
            "parent_recid": parent_recid,
        },
    }


def generate_download_stats(analytics, cache):
    """Generator for download statistics actions."""
    for analytic in analytics:
        if record_exists(analytic, cache) and file_key_exists(analytic, cache):
            yield to_download(analytic, cache)


def to_view(entry, cache):
    """To view."""
    recid = entry.pid
    year_month_day = entry.year_month_day
    event_name = "record-view"
    year_month = year_month_day.rsplit("-", 1)[0]
    count = entry.views
    unique_count = entry.visits
    parent_recid = cache.get_parent_pid(recid)

    return {
        "_id": f"ui_{recid}-{year_month_day}",
        "_index": f"stats-{event_name}-{year_month}",
        "_source": {
            # Since those entries are synthetic anyway, we place them at
            # the start of the day
            "timestamp": f"{year_month_day}T00:00:00",
            "unique_id": f"ui_{recid}",
            "count": count,
            "updated_timestamp": datetime.now(timezone.utc).isoformat(),
            "unique_count": unique_count,
            "recid": recid,
            "parent_recid": parent_recid,
            "via_api": False,
        },
    }


def generate_view_stats(analytics, cache):
    """Generator for view statistics actions."""
    for analytic in analytics:
        if record_exists(analytic, cache):
            yield to_view(analytic, cache)


def ingest_statistics(actions):
    """Ingest statistics actions."""
    search.helpers.bulk(
        current_search_client,
        actions,
        stats_only=True,
        chunk_size=50,
    )



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Northwestern University.
#
# invenio-analytics-importer is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Command-line interface."""

import asyncio
from pathlib import Path

import click
import flask

from invenio_analytics_importer.cache import (
    fill_downloads_cache,
    fill_views_cache,
)
from invenio_analytics_importer.convert import (
    generate_download_analytics,
    generate_view_analytics,
)
from invenio_analytics_importer.ingest import (
    generate_download_stats,
    generate_view_stats,
    ingest_statistics,
)
from invenio_analytics_importer.read import read_raw_analytics_from_filepaths
from invenio_analytics_importer.retrieve import (
    retrieve_period_analytics,
)


@click.group()
@flask.cli.with_appcontext
def analytics():
    """Analytics importer commands."""


@analytics.command()
@click.option(
    "--views",
    "kind",
    flag_value="views",
    default="views",
    help="Retrieve views.",
)
@click.option(
    "--downloads",
    "kind",
    flag_value="downloads",
    help="Retrieve downloads.",
)
@click.option(
    "--from",
    "-f",
    "year_month_from",
    required=True,
    help="Download analytics from this month (inclusive).",
)
@click.option(
    "--to",
    "-t",
    "year_month_to",
    required=True,
    help="Download analytics to this month (inclusive).",
)
@click.option(
    "--output-dir",
    "-o",
    type=click.Path(exists=True, path_type=Path),
    required=True,
    help="Place resulting analytics file in this directory.",
)
def retrieve(kind, year_month_from, year_month_to, output_dir):
    """Retrieve analytics from the web.

    For 'kind' the last --views/--downloads takes precedence.
    """
    asyncio.run(
        retrieve_period_analytics(
            provider="matomo",  # hardcoded for now
            kind=kind,
            period=(year_month_from, year_month_to),
            output_dir=output_dir,
        )
    )


@analytics.command()
@click.option(
    "--views",
    "kind",
    flag_value="views",
    default="views",
    help="Retrieve views.",
)
@click.option(
    "--downloads",
    "kind",
    flag_value="downloads",
    help="Retrieve downloads.",
)
@click.option(
    "-f", "--filepath", type=click.Path(exists=True, path_type=Path),
    multiple=True
)
def ingest(kind, filepath):
    """Ingest stats from given filepaths into your RDM instance."""
    # filepath is actually a list of filepaths
    filepaths = filepath

    # Raw read is same across views/downloads
    raw_analytics = read_raw_analytics_from_filepaths(filepaths)

    # Convert raw analytics to processed analytics
    if kind == "views":
        iter_analytics = generate_view_analytics(raw_analytics)
    elif kind == "downloads":
        iter_analytics = generate_download_analytics(raw_analytics)
    else:
        exit(1)

    # Prepare cache
    raw_analytics_for_cache = read_raw_analytics_from_filepaths(filepaths)
    if kind == "views":
        iter_analytics_for_cache = generate_view_analytics(
            raw_analytics_for_cache
        )
        cache = fill_views_cache(iter_analytics_for_cache)
    elif kind == "downloads":
        iter_analytics_for_cache = generate_download_analytics(
            raw_analytics_for_cache
        )
        cache = fill_downloads_cache(iter_analytics_for_cache)

    # Convert to ingestable
    if kind == "views":
        stats_for_ingest = generate_view_stats(iter_analytics, cache)
    elif kind == "downloads":
        stats_for_ingest = generate_download_stats(iter_analytics, cache)

    # ingest is same across kind
    ingest_statistics(stats_for_ingest)
